@@ -1,6 +1,6 @@
 /*
   Racket
-  Copyright (c) 2004-2012 PLT Scheme Inc.
+  Copyright (c) 2004-2013 PLT Scheme Inc.
   Copyright (c) 1995-2001 Matthew Flatt
   All rights reserved.
 
@@ -110,6 +110,9 @@ typedef long FILE;
 #ifndef MSC_IZE
 # define MSC_IZE(x) x
 #endif
+#ifndef M_MSC_IZE
+# define M_MSC_IZE(x) x
+#endif
 #ifndef MSCBOR_IZE
 # define MSCBOR_IZE(x) MSC_IZE(x)
 #endif
@@ -127,10 +130,14 @@ typedef long FILE;
    old way. */
 #ifdef MZ_USE_OLD_ARRAY_STYLE
 # define mzFLEX_ARRAY_DECL 1
+# define mzFLEX_ARRAY4_DECL 4
 # define mzFLEX_DELTA 1
+# define mzFLEX4_DELTA 4
 #else
 # define mzFLEX_ARRAY_DECL /* empty */
+# define mzFLEX_ARRAY4_DECL /* empty */
 # define mzFLEX_DELTA 0
+# define mzFLEX4_DELTA 0
 #endif
 
 #ifdef MZ_XFORM
@@ -237,8 +244,8 @@ typedef _int64 mzlonglong;
 typedef unsigned _int64 umzlonglong;
 #else
 # if defined(NO_LONG_LONG_TYPE) || defined(SIXTY_FOUR_BIT_INTEGERS)
-typedef long mzlonglong;
-typedef unsigned long umzlonglong;
+typedef intptr_t mzlonglong;
+typedef uintptr_t umzlonglong;
 # else
 typedef long long mzlonglong;
 typedef unsigned long long umzlonglong;
@@ -331,7 +338,7 @@ typedef struct {
 typedef struct Scheme_Symbol {
   Scheme_Inclhash_Object iso; /* 1 in low bit of keyex indicates uninterned */
   intptr_t len;
-  char s[4]; /* Really, a number of chars to match `len' */
+  char s[mzFLEX_ARRAY4_DECL];
 } Scheme_Symbol;
 
 typedef struct Scheme_Vector {
@@ -660,16 +667,14 @@ typedef struct Scheme_Offset_Cptr
    Do not use them directly. */
 #define SCHEME_PRIM_OPT_MASK (1 | 2)
 #define SCHEME_PRIM_IS_PRIMITIVE 4
-#define SCHEME_PRIM_IS_UNSAFE_OMITABLE 8
-#define SCHEME_PRIM_IS_OMITABLE 16
+#define SCHEME_PRIM_IS_MULTI_RESULT 8
+#define SCHEME_PRIM_IS_CLOSURE 16
 #define SCHEME_PRIM_OTHER_TYPE_MASK (32 | 64 | 128 | 256)
-#define SCHEME_PRIM_IS_MULTI_RESULT 512
-#define SCHEME_PRIM_IS_BINARY_INLINED 1024
-#define SCHEME_PRIM_IS_UNSAFE_FUNCTIONAL 2048
-#define SCHEME_PRIM_IS_METHOD 4096
-#define SCHEME_PRIM_IS_CLOSURE 8192
-#define SCHEME_PRIM_IS_UNARY_INLINED 16384
-#define SCHEME_PRIM_IS_NARY_INLINED 32768
+#define SCHEME_PRIM_IS_METHOD 512
+
+#define SCHEME_PRIM_OPT_INDEX_SIZE 6
+#define SCHEME_PRIM_OPT_INDEX_SHIFT 10
+#define SCHEME_PRIM_OPT_INDEX_MASK ((1 << SCHEME_PRIM_OPT_INDEX_SIZE) - 1)
 
 /* Values with SCHEME_PRIM_OPT_MASK, earlier implies later: */
 #define SCHEME_PRIM_OPT_FOLDING    3
@@ -679,8 +684,10 @@ typedef struct Scheme_Offset_Cptr
 /* Values with SCHEME_PRIM_OTHER_TYPE_MASK */
 #define SCHEME_PRIM_STRUCT_TYPE_INDEXLESS_GETTER (32 | 256)
 #define SCHEME_PRIM_STRUCT_TYPE_CONSTR           128
+#define SCHEME_PRIM_STRUCT_TYPE_SIMPLE_CONSTR    (32 | 64 | 128)
 #define SCHEME_PRIM_STRUCT_TYPE_INDEXLESS_SETTER 256
 #define SCHEME_PRIM_STRUCT_TYPE_INDEXED_SETTER   (128 | 256)
+#define SCHEME_PRIM_STRUCT_TYPE_BROKEN_INDEXED_SETTER   (32 | 128)
 #define SCHEME_PRIM_TYPE_PARAMETER               64
 #define SCHEME_PRIM_TYPE_STRUCT_PROP_GETTER      (64 | 128)
 #define SCHEME_PRIM_SOMETIMES_INLINED            (64 | 256)
@@ -1046,6 +1053,8 @@ typedef struct Scheme_Thread {
   Scheme_Object *dead_box;      /* contains non-zero when the thread is dead */
   Scheme_Object *running_box;   /* contains pointer to thread when it's running */
 
+  struct Scheme_Thread *gc_prep_chain;
+
   struct Scheme_Thread *nester, *nestee;
 
   struct future_t *current_ft;
@@ -1073,7 +1082,7 @@ typedef struct Scheme_Thread {
 
   struct Scheme_Marshal_Tables *current_mt;
 
-  Scheme_Object *constant_folding; /* compiler hack */
+  struct Optimize_Info *constant_folding; /* compiler hack */
   Scheme_Object *reading_delayed; /* reader hack */
 
   Scheme_Object *(*overflow_k)(void);
@@ -1278,6 +1287,7 @@ enum {
   MZCONFIG_CODE_INSPECTOR,
 
   MZCONFIG_USE_COMPILED_KIND,
+  MZCONFIG_USE_COMPILED_ROOTS,
   MZCONFIG_USE_USER_PATHS,
   MZCONFIG_USE_LINK_PATHS,
 
@@ -1376,6 +1386,7 @@ struct Scheme_Port
   Scheme_Location_Fun location_fun;
   Scheme_Count_Lines_Fun count_lines_fun;
   Scheme_Buffer_Mode_Fun buffer_mode_fun;
+  Scheme_Object *position_redirect; /* for `file-position' */
 };
 
 struct Scheme_Input_Port
@@ -1765,6 +1776,7 @@ MZ_EXTERN void scheme_set_startup_load_on_demand(int);
 MZ_EXTERN void scheme_set_ignore_user_paths(int);
 MZ_EXTERN void scheme_set_ignore_link_paths(int);
 MZ_EXTERN void scheme_set_logging(int syslog_level, int stderr_level);
+MZ_EXTERN void scheme_set_logging_spec(Scheme_Object *syslog_level, Scheme_Object *stderr_level);
 
 MZ_EXTERN int scheme_get_allow_set_undefined();
 
@@ -1831,9 +1843,11 @@ MZ_EXTERN void scheme_set_addon_dir(Scheme_Object *p);
 MZ_EXTERN void scheme_set_links_file(Scheme_Object *p);
 MZ_EXTERN void scheme_set_command_line_arguments(Scheme_Object *vec);
 MZ_EXTERN void scheme_set_compiled_file_paths(Scheme_Object *list);
+MZ_EXTERN void scheme_set_compiled_file_roots(Scheme_Object *list);
 
 MZ_EXTERN void scheme_init_collection_paths(Scheme_Env *global_env, Scheme_Object *extra_dirs);
 MZ_EXTERN void scheme_init_collection_paths_post(Scheme_Env *global_env, Scheme_Object *extra_dirs, Scheme_Object *extra_post_dirs);
+MZ_EXTERN void scheme_init_compiled_roots(Scheme_Env *global_env, const char *paths);
 
 MZ_EXTERN void scheme_seal_parameters();
 
@@ -1892,7 +1906,7 @@ MZ_EXTERN Scheme_Object *scheme_register_parameter(Scheme_Prim *function, char *
 /*                              addrinfo                                  */
 /*========================================================================*/
 
-#ifdef HAVE_GETADDRINFO
+#if defined(HAVE_GETADDRINFO) || defined(__MINGW32__)
 # define mz_addrinfo addrinfo
 #else
 struct mz_addrinfo {
