@@ -27,8 +27,8 @@ extern "C" {
 
 #if defined(MZ_USE_PLACES) || defined(MZ_USE_FUTURES)
 # define USE_THREAD_LOCAL
-# if _MSC_VER
-#  ifdef _WIN64
+# ifdef _WIN32
+#  if defined(_WIN64) && !defined(__MINGW32__)
 #   define THREAD_LOCAL __declspec(thread)
 #   define MZ_THREAD_EXTERN extern
 #   define IMPLEMENT_THREAD_LOCAL_EXTERNALLY_VIA_PROC
@@ -54,7 +54,7 @@ extern "C" {
     && !defined(SCHEME_EMBEDDED_NO_DLL)
 # define MZ_DLLIMPORT __declspec(dllimport)
 # define MZ_DLLEXPORT __declspec(dllexport)
-# ifdef __mzscheme_private__
+# if defined(__mzscheme_private__) || defined(__MINGW32_DELAY_LOAD__)
 #  define MZ_DLLSPEC __declspec(dllexport)
 # else
 #  define MZ_DLLSPEC __declspec(dllimport)
@@ -230,6 +230,7 @@ typedef struct Thread_Local_Variables {
   struct Scheme_Thread *scheme_current_thread_;
   struct Scheme_Thread *scheme_main_thread_;
   struct Scheme_Thread *scheme_first_thread_;
+  struct Scheme_Thread *gc_prep_thread_chain_;
   struct Scheme_Thread_Set *scheme_thread_set_top_;
   struct Scheme_Current_LWC *scheme_current_lwc_;
   int num_running_threads_;
@@ -272,6 +273,8 @@ typedef struct Thread_Local_Variables {
   struct gmp_tmp_stack gmp_tmp_xxx_;
   struct gmp_tmp_stack *gmp_tmp_current_;
   struct Scheme_Logger *scheme_main_logger_;
+  struct Scheme_Logger *scheme_gc_logger_;
+  struct Scheme_Logger *scheme_future_logger_;
   int intdef_counter_;
   int builtin_ref_counter_;
   int env_uid_counter_;
@@ -333,6 +336,7 @@ typedef struct Thread_Local_Variables {
   int place_evts_array_size_;
   struct Evt **place_evts_;
   struct Scheme_Place_Object *place_object_;
+  struct Scheme_Place *all_child_places_;
   struct Scheme_Object **reusable_ifs_stack_;
   struct Scheme_Object *empty_self_shift_cache_;
   struct Scheme_Bucket_Table *scheme_module_code_cache_;
@@ -414,6 +418,31 @@ MZ_EXTERN uintptr_t scheme_tls_delta;
 MZ_EXTERN int scheme_tls_index;
 #  endif
 static __inline Thread_Local_Variables **scheme_get_thread_local_variables_ptr(void) {
+# ifdef __MINGW32__
+  Thread_Local_Variables **x;
+#  ifdef _WIN64
+  asm (
+       "mov %%gs:(0x58), %%rax;"
+       "mov (%%rax), %%rax;"
+       "add %1, %%rax;"
+       "mov %%rax, %0;"
+       :"=r"(x)        /* output */
+       :"r"(scheme_tls_delta)
+       :"%rax"  /* clobbered register */
+       );
+#  else
+  asm (
+       "mov %%fs:(0x2C), %%eax;"
+       "mov (%%eax), %%eax;"
+       "add %1, %%eax;"
+       "mov %%eax, %0;"
+       :"=r"(x)        /* output */
+       :"r"(scheme_tls_delta)
+       :"%eax"  /* clobbered register */
+       );
+#  endif
+  return x;
+# else
   __asm { mov eax, FS:[0x2C]
 #  ifdef MZ_USE_WIN_TLS_VIA_DLL
           add eax, scheme_tls_index
@@ -421,6 +450,7 @@ static __inline Thread_Local_Variables **scheme_get_thread_local_variables_ptr(v
           mov eax, [eax]
           add eax, scheme_tls_delta }
   /* result is in eax */
+# endif
 }
 static __inline Thread_Local_Variables *scheme_get_thread_local_variables(void) {
   return *scheme_get_thread_local_variables_ptr();
@@ -575,6 +605,7 @@ XFORM_GC_VARIABLE_STACK_THROUGH_THREAD_LOCAL;
 #define scheme_current_thread XOA (scheme_get_thread_local_variables()->scheme_current_thread_)
 #define scheme_main_thread XOA (scheme_get_thread_local_variables()->scheme_main_thread_)
 #define scheme_first_thread XOA (scheme_get_thread_local_variables()->scheme_first_thread_)
+#define gc_prep_thread_chain XOA (scheme_get_thread_local_variables()->gc_prep_thread_chain_)
 #define scheme_thread_set_top XOA (scheme_get_thread_local_variables()->scheme_thread_set_top_)
 #define scheme_current_lwc XOA (scheme_get_thread_local_variables()->scheme_current_lwc_)
 #define num_running_threads XOA (scheme_get_thread_local_variables()->num_running_threads_)
@@ -617,6 +648,8 @@ XFORM_GC_VARIABLE_STACK_THROUGH_THREAD_LOCAL;
 #define gmp_tmp_xxx XOA (scheme_get_thread_local_variables()->gmp_tmp_xxx_)
 #define gmp_tmp_current XOA (scheme_get_thread_local_variables()->gmp_tmp_current_)
 #define scheme_main_logger XOA (scheme_get_thread_local_variables()->scheme_main_logger_)
+#define scheme_gc_logger XOA (scheme_get_thread_local_variables()->scheme_gc_logger_)
+#define scheme_future_logger XOA (scheme_get_thread_local_variables()->scheme_future_logger_)
 #define intdef_counter XOA (scheme_get_thread_local_variables()->intdef_counter_)
 #define builtin_ref_counter XOA (scheme_get_thread_local_variables()->builtin_ref_counter_)
 #define env_uid_counter XOA (scheme_get_thread_local_variables()->env_uid_counter_)
@@ -678,6 +711,7 @@ XFORM_GC_VARIABLE_STACK_THROUGH_THREAD_LOCAL;
 #define place_evts_array_size XOA (scheme_get_thread_local_variables()->place_evts_array_size_)
 #define place_evts XOA (scheme_get_thread_local_variables()->place_evts_)
 #define place_object XOA (scheme_get_thread_local_variables()->place_object_)
+#define all_child_places XOA (scheme_get_thread_local_variables()->all_child_places_)
 #define reusable_ifs_stack XOA (scheme_get_thread_local_variables()->reusable_ifs_stack_)
 #define empty_self_shift_cache XOA (scheme_get_thread_local_variables()->empty_self_shift_cache_)
 #define scheme_module_code_cache XOA (scheme_get_thread_local_variables()->scheme_module_code_cache_)
