@@ -82,7 +82,6 @@
 # include "uconfig.h"
 
 # define USE_EXPLICT_FP_FORM_CHECK
-# define POW_HANDLES_INF_CORRECTLY
 
 # include <errno.h>
 # ifdef ECHRNG
@@ -144,8 +143,6 @@
 
 # define SELECT_INCLUDE
 
-# define POW_HANDLES_INF_CORRECTLY
-
 # define USE_FCNTL_O_NONBLOCK
 
 # define USE_TIMEZONE_VAR_W_DLS
@@ -261,7 +258,6 @@
 #endif
 
 # define USE_IEEE_FP_PREDS
-# define POW_HANDLES_INF_CORRECTLY
 
 # define USE_DYNAMIC_FDSET_SIZE
 
@@ -301,6 +297,10 @@
 #  define SCHEME_PLATFORM_LIBRARY_SUBPATH "i386-openbsd"
 # endif
 
+/* Same reason as for FreeBSD? */
+# define ASSUME_FIXED_STACK_SIZE
+# define FIXED_STACK_SIZE 1048576
+
 # include "uconfig.h"
 # undef HAS_STANDARD_IOB
 
@@ -311,7 +311,6 @@
 #endif
 
 # define USE_IEEE_FP_PREDS
-# define POW_HANDLES_INF_CORRECTLY
 
 # define USE_DYNAMIC_FDSET_SIZE
 
@@ -365,7 +364,7 @@
 #  error Unported platform.
 # endif
 
-/* pthreads always enabled via configure', and
+/* pthreads always enabled via `configure', and
    initial pthread's stack size doesn't use rlimit: */
 # define ASSUME_FIXED_STACK_SIZE
 # define FIXED_STACK_SIZE 1048576
@@ -382,9 +381,6 @@
 # define USE_UNDERSCORE_SETJMP
 
 # define USE_IEEE_FP_PREDS
-# ifndef ASM_DBLPREC_CONTROL_87
-#  define POW_HANDLES_INF_CORRECTLY
-# endif
 
 # define USE_DYNAMIC_FDSET_SIZE
 
@@ -520,7 +516,7 @@
   /* See the "worksp" directory for more MSVC details.       */
 
 #if (defined(__BORLANDC__) \
-     || (defined(_MSC_VER) \
+     || ((defined(_MSC_VER) || defined(__MINGW32__)) \
          && (defined(__WIN32__) || defined(WIN32) || defined(_WIN32))))
 
 # ifdef _WIN64
@@ -531,7 +527,7 @@
 
 # define SYSTEM_TYPE_NAME "windows"
 # define DOS_FILE_SYSTEM
-# if defined(_MSC_VER)
+# if defined(_MSC_VER) || defined(__MINGW32__)
 #  define NO_READDIR
 #  define USE_FINDFIRST
 #  define NO_READLINK
@@ -589,6 +585,10 @@
 # define TRIG_ZERO_NEEDS_SIGN_CHECK
 # define NEED_TO_DEFINE_MATHERR
 #endif
+#ifdef __MINGW32__
+# define USE_IEEE_FP_PREDS
+# define ATAN2_DOESNT_WORK_WITH_INFINITIES
+#endif
 
 # define IO_INCLUDE
 # define DONT_IGNORE_PIPE_SIGNAL
@@ -603,6 +603,7 @@
 
 # define SIGSET_IS_SIGNAL
 # define SIGSET_NEEDS_REINSTALL
+# define NO_SIGHUP_HANDLER
 
 #define USE_WINSOCK_TCP
 
@@ -611,7 +612,14 @@
 /* MS Visual C++ likes underscore prefixes */
 #if defined(_MSC_VER)
 # define MSC_IZE(x) _ ## x
+# define M_MSC_IZE(x) MSC_IZE(x)
 # define DIRECT_INCLUDE
+# else
+# define M_MSC_IZE(x) x
+#endif
+
+#if defined(__MINGW32__)
+# define MSC_IZE(x) _ ## x
 #endif
 
 #ifdef __BORLANDC__
@@ -623,6 +631,9 @@
 #if defined(_MSC_VER)
 # define IGNORE_BY_MS_CONTROL_87
 #endif
+#if defined(__MINGW32__)
+# define ASM_DBLPREC_CONTROL_87
+#endif
 
 # define REGISTER_POOR_MACHINE
 
@@ -773,6 +784,7 @@
 #ifdef MZ_NO_JIT_SSE
 # define ASM_DBLPREC_CONTROL_87
 #endif
+# define POW_HANDLES_CASES_CORRECTLY
 
 # define MZ_JIT_USE_MPROTECT
 
@@ -861,7 +873,6 @@
 # define NO_STAT_PROC
 # define DONT_IGNORE_PIPE_SIGNAL
 
-# define POW_HANDLES_INF_CORRECTLY
 # define TRIG_ZERO_NEEDS_SIGN_CHECK
 
 # define MACOS_UNICODE_SUPPORT
@@ -947,7 +958,6 @@
 # define DONT_IGNORE_PIPE_SIGNAL
 # define DONT_IGNORE_FPE_SIGNAL
 
-# define POW_HANDLES_INF_CORRECTLY
 # define USE_PALM_INF_TESTS
 
 # define FLAGS_ALREADY_SET
@@ -973,6 +983,8 @@
 
 # define FLAGS_ALREADY_SET
 
+# define USE_DYNAMIC_FDSET_SIZE
+
 #if defined(i386)
 # define MZ_USE_JIT_I386
 # define MZ_JIT_USE_MPROTECT
@@ -983,6 +995,58 @@
 # define MZ_USE_DWARF_LIBUNWIND
 #endif
 
+#endif
+
+  /************ Dragonfly *************/
+
+#if defined(__DragonFly__)
+ 
+# if defined(__i386__)
+#  define SCHEME_PLATFORM_LIBRARY_SUBPATH "i386-dragonfly"
+#  define REGISTER_POOR_MACHINE
+#  define MZ_USE_JIT_I386
+#  ifndef MZ_JIT_X86_SSE
+#     define ASM_DBLPREC_CONTROL_87
+#  endif
+# elif defined(__amd64__)
+#  define SCHEME_PLATFORM_LIBRARY_SUBPATH "amd64-dragonfly"
+#  define REGISTER_POOR_MACHINE
+#  define MZ_USE_JIT_X86_64
+#  ifdef MZ_NO_JIT_SSE
+#     define ASM_DBLPREC_CONTROL_87
+#  endif
+# else
+#  error Unported platform.
+# endif
+
+/* pthreads always enabled via `configure', and
+   initial pthread's stack size doesn't use rlimit: */
+# define ASSUME_FIXED_STACK_SIZE
+# define FIXED_STACK_SIZE 1048576
+
+# include "uconfig.h"
+# undef HAS_STANDARD_IOB
+
+# define HAS_BSD_IOB
+
+# define USE_UNDERSCORE_SETJMP
+
+# define USE_IEEE_FP_PREDS
+# ifndef ASM_DBLPREC_CONTROL_87
+#  define POW_HANDLES_INF_CORRECTLY
+# endif
+
+# define USE_DYNAMIC_FDSET_SIZE
+
+# define SIGSET_IS_SIGNAL
+
+# define USE_TM_GMTOFF_FIELD
+# define USE_TM_ZONE_FIELD
+
+# define MZ_JIT_USE_MPROTECT
+
+# define FLAGS_ALREADY_SET
+
 #endif
 
   /***************************************************/
@@ -1323,9 +1387,9 @@
     SunOS/Solaris, and HP/UX by explicit pre-checking the form of the 
     number and looking for values that are obviously +inf.0 or -inf.0 */
 
- /* POW_HANDLES_INF_CORRECTLY indicates that thw pow() library procedure
-    handles +/-inf.0 correctly. Otherwise, code in inserted to specifically
-    check for infinite arguments. */
+ /* POW_HANDLES_CASES_CORRECTLY indicates that the pow() library procedure
+    handles all +/-inf.0, +/-0.0, or +nan.0 cases according to C99. This
+    might save time on redundant checking before Racket calls pow(). */
     
  /* ATAN2_DOESNT_WORK_WITH_INFINITIES indicates that atan2(+/-inf, +/-inf)
     is not the same as atan2(1, 1). */ 
@@ -1508,7 +1572,12 @@
  /* NO_INLINE_KEYWORD indicates that the C compiler doesn't recognize
     C's `inline' keyword. */
 
- /* NO_USER_BREAK_HANDLER turns off handling of INT signal in main.c */
+ /* NO_USER_BREAK_HANDLER turns off handling of SIGINT, SIGTERM, and 
+    SIGHUP in main.c */
+
+ /* NO_SIGTERM_HANDLER turns off handling of SIGTERM in main.c */
+
+ /* NO_SIGHUP_HANDLER turns off handling of SIGHUP in main.c */
 
  /* DIR_INCLUDE if there's a <dir.h> file (mainly for Windows). */
 
