@@ -1,6 +1,6 @@
 /*
   Racket
-  Copyright (c) 2004-2012 PLT Scheme Inc.
+  Copyright (c) 2004-2013 PLT Scheme Inc.
   Copyright (c) 1995-2001 Matthew Flatt
 
     This library is free software; you can redistribute it and/or
@@ -280,6 +280,10 @@ Scheme_Env *scheme_engine_instance_init()
 #endif
   make_kernel_env();
 
+  scheme_init_logging_once();
+
+  scheme_init_compenv_symbol();
+
 #if defined(MZ_PLACES_WAITPID)
   scheme_places_start_child_signal_handler();
 #endif
@@ -764,7 +768,7 @@ void scheme_prepare_env_renames(Scheme_Env *env, int kind)
   if (!env->rename_set) {
     Scheme_Object *rns, *insp;
 
-    insp = env->insp;
+    insp = env->access_insp;
     if (!insp)
       insp = scheme_get_param(scheme_current_config(), MZCONFIG_CODE_INSPECTOR);
 
@@ -904,7 +908,8 @@ void scheme_prepare_exp_env(Scheme_Env *env)
     eenv->module = env->module;
     eenv->module_registry = env->module_registry;
     eenv->module_pre_registry = env->module_pre_registry;
-    eenv->insp = env->insp;
+    eenv->access_insp = env->access_insp;
+    eenv->guard_insp = env->guard_insp;
 
     modchain = SCHEME_VEC_ELS(env->modchain)[1];
     if (SCHEME_FALSEP(modchain)) {
@@ -946,7 +951,8 @@ void scheme_prepare_template_env(Scheme_Env *env)
     eenv->module = env->module;
     eenv->module_registry = env->module_registry;
     eenv->module_pre_registry = env->module_pre_registry;
-    eenv->insp = env->insp;
+    eenv->guard_insp = env->guard_insp;
+    eenv->access_insp = env->access_insp;
 
     modchain = SCHEME_VEC_ELS(env->modchain)[2];
     if (SCHEME_FALSEP(modchain)) {
@@ -987,7 +993,8 @@ void scheme_prepare_label_env(Scheme_Env *env)
     lenv->module = env->module;
     lenv->module_registry = env->module_registry;
     lenv->module_pre_registry = env->module_pre_registry;
-    lenv->insp = env->insp;
+    lenv->guard_insp = env->guard_insp;
+    lenv->access_insp = env->access_insp;
 
     modchain = scheme_make_vector(5, scheme_false);    
     prev_modules = scheme_make_hash_table(SCHEME_hash_ptr);
@@ -1020,7 +1027,8 @@ Scheme_Env *scheme_copy_module_env(Scheme_Env *menv, Scheme_Env *ns, Scheme_Obje
   menv2->module = menv->module;
   menv2->module_registry = ns->module_registry;
   menv2->module_pre_registry = ns->module_pre_registry;
-  menv2->insp = menv->insp;
+  menv2->guard_insp = menv->guard_insp;
+  menv2->access_insp = menv->access_insp;
 
   menv2->instance_env = menv2;
 
@@ -1584,7 +1592,7 @@ namespace_variable_value(int argc, Scheme_Object *argv[])
   use_map = ((argc > 1) ? SCHEME_TRUEP(argv[1]) : 1);
   if ((argc > 2) && SCHEME_TRUEP(argv[2])
       && !scheme_check_proc_arity(NULL, 0, 2, argc, argv))
-    scheme_wrong_contract("namespace-variable-value", "(or/c (-> any) #f)", 1, argc, argv);
+    scheme_wrong_contract("namespace-variable-value", "(or/c (-> any) #f)", 2, argc, argv);
   if ((argc > 3) && !SCHEME_NAMESPACEP(argv[3]))
     scheme_wrong_contract("namespace-variable-value", "namespace?", 3, argc, argv);
 
@@ -1762,7 +1770,7 @@ static Scheme_Object *do_variable_namespace(const char *who, int tl, int argc, S
                             "variable reference", 1, v,
                             NULL);
     }
-    return env->module->insp;
+    return env->access_insp;
   } else if (tl) {
     /* return env directly; need to set up  */
     if (!env->phase && env->module)
@@ -2272,13 +2280,19 @@ local_get_shadower(int argc, Scheme_Object *argv[])
   uid = scheme_find_local_shadower(sym, sym_marks, env);
 
   if (!uid) {
-    /* No lexical shadower, but strip module context, if any */
-    sym = scheme_stx_strip_module_context(sym);
-    /* Add current module context, if any */
-    sym = local_module_introduce(1, &sym);
+    uid = scheme_tl_id_sym(env->genv, sym, NULL, 0, 
+                           scheme_make_integer(env->genv->phase), NULL);
+    if (!SAME_OBJ(uid, SCHEME_STX_VAL(sym))) {
+      /* has a toplevel biding via marks or context; keep it */
+    } else {
+      /* No lexical shadower, but strip module context, if any */
+      sym = scheme_stx_strip_module_context(sym);
+      /* Add current module context, if any */
+      sym = local_module_introduce(1, &sym);
 
-    if (!scheme_stx_is_clean(orig_sym))
-      sym = scheme_stx_taint(sym);
+      if (!scheme_stx_is_clean(orig_sym))
+        sym = scheme_stx_taint(sym);
+    }
 
     return sym;
   }
@@ -2435,7 +2449,7 @@ Scheme_Object *scheme_get_local_inspector()
   Scheme_Thread *p = scheme_current_thread;
 
   if (p->current_local_menv)
-    return p->current_local_menv->module->insp;
+    return p->current_local_menv->access_insp;
   else
     return scheme_get_param(scheme_current_config(), MZCONFIG_CODE_INSPECTOR);
 }
