@@ -1,6 +1,6 @@
 /*
   Racket
-  Copyright (c) 2004-2012 PLT Scheme Inc.
+  Copyright (c) 2004-2013 PLT Scheme Inc.
   Copyright (c) 2000-2001 Matthew Flatt
 
     This library is free software; you can redistribute it and/or
@@ -500,7 +500,7 @@ Scheme_Object *scheme_read_number(const mzchar *str, intptr_t len,
   }
 
   /* look for +inf.0, etc: */
-  if (len -delta == 6) {
+  if (len - delta == 6) {
     Scheme_Object *special;
     special = read_special_number(str, delta);
     if (special) {
@@ -525,7 +525,7 @@ Scheme_Object *scheme_read_number(const mzchar *str, intptr_t len,
       s2 = (mzchar *)scheme_malloc_atomic((len - delta - 6 + 4 + 1) * sizeof(mzchar));
       s2[0] = '+';
       s2[1] = '0';
-      s2[2] = '.';
+      s2[2] = (SCHEME_DBLP(special) ? '.' : 's');
       s2[3] = '0';
       memcpy(s2 + 4, str + delta + 6, (len - delta - 5) * sizeof(mzchar));
     } else {
@@ -536,7 +536,7 @@ Scheme_Object *scheme_read_number(const mzchar *str, intptr_t len,
 	memcpy(s2, str + delta, (len - delta - 7) * sizeof(mzchar));
 	s2[len - delta - 7] = '+';
 	s2[len - delta - 7 + 1] = '0';
-	s2[len - delta - 7 + 2] = '.';
+	s2[len - delta - 7 + 2] = (SCHEME_DBLP(special) ? '.' : 's');
 	s2[len - delta - 7 + 3] = '0';
 	s2[len - delta - 7 + 4] = 'i';
 	s2[len - delta - 7 + 5] = 0;
@@ -1531,12 +1531,8 @@ string_to_number (int argc, Scheme_Object *argv[])
   return v;
 }
 
-
-static char *double_to_string (double d, int alloc, int was_single)
+char *scheme_double_to_string (double d, char* s, int slen, int was_single, int *used_buffer)
 {
-  char buffer[100], *s;
-  int l, i, digits;
-
   if (MZ_IS_NAN(d))
 #ifdef MZ_USE_SINGLE_FLOATS
     if (was_single)
@@ -1581,13 +1577,15 @@ static char *double_to_string (double d, int alloc, int was_single)
        (single). That's big enough to get most right, small enough to
        avoid nonsense digits. But we'll loop in case it's not precise
        enough to get read-write invariance: */
+    int i, l, digits;
     GC_CAN_IGNORE char *loc;
+    char *buffer = s;
     if (was_single)
       digits = 6;
     else
       digits = 14;
     loc = scheme_push_c_numeric_locale();
-    while (digits < 30) {
+    while (digits < 30 && digits < slen) {
       double check;
       GC_CAN_IGNORE char *ptr;
 
@@ -1637,7 +1635,20 @@ static char *double_to_string (double d, int alloc, int was_single)
       }
     }
 #endif
-    
+    *used_buffer = 1;
+  }
+
+  return s;
+}
+
+static char *double_to_string (double d, int alloc, int was_single)
+{
+  char buffer[100];
+  char *s;
+  int used_buffer = 0;
+  s = scheme_double_to_string(d, buffer, 100, was_single, &used_buffer);
+
+  if (used_buffer) {
     s = (char *)scheme_malloc_atomic(strlen(buffer) + 1);
     strcpy(s, buffer);
     alloc = 0;
@@ -1645,6 +1656,7 @@ static char *double_to_string (double d, int alloc, int was_single)
 
   if (alloc) {
     char *s2;
+    int l;
     l = strlen(s) + 1;
     s2 = (char *)scheme_malloc_atomic(l);
     memcpy(s2, s, l);
